/* zkl_yajl : front end to yajl
 * This file a munged json_reformat.c
 * 
 * Compile multi-threaded (MSVC properties/C/Code Generation/Runtime library)
 * Linked against zkl.lib (zkl exports)
 *    msvcrt.lib.
 *    In MSVC properties/Linker/Input, set "Ignore All Default Libraries" to
 *    "Yes (/NODEFAULTLIB)". This creates a much smaller DLL and is why you
 *    need to link against msvcrt.lib
 * 
 * NOT thread safe
 * 
 * Copyright (c) 2015 Craig Durland craigd@zenkinetic.com
 * License: The zlib/libpng License,
 * 	http://www.opensource.org/licenses/zlib-license.php
 */

//#include <stdio.h>
#include <string.h>
//#include <stdlib.h>

#include "zklObject.h"	
#include "zklImports.h"
#include "zklClass.h"
#include "zklData.h"
#include "zklDictionary.h"
#include "zklList.h"
#include "zklMemory.h"
#include "zklMethod.h"
#include "zklString.h"
#include "zklUtil.h"

#include <yajl/yajl_parse.h>

#define MAX_STACK 200
typedef struct 
{
   Instance  instance;
   Instance *root, *node, *key, *src, *stack[MAX_STACK];
   int nodeType, n, closed;
   pVM vm;
   yajl_handle hand;
} ZY_CTX;

static ZKL_Object yajlObject;

static int yajlFree(Instance *self)
{
   ZY_CTX *ctx = (ZY_CTX *)self;
   if(ctx->closed) return 1;
   yajl_free(ctx->hand);
   ctx->closed = 1;

   return 1; 
}

static void yajlMarker(Instance *self)
{
   ZY_CTX *ctx = (ZY_CTX *)self;
   if(!ctx->closed)
   {
      instanceMark(ctx->root); instanceMark(ctx->key); 
      if(ctx->src) instanceMark(ctx->src);
   }
}

static int append(void *_ctx,Instance *i)
{
   ZY_CTX *ctx = (ZY_CTX *)_ctx;
   if(ctx->closed)
   {
      vmThrow(ctx->vm,E_IO_ERROR,"yajl is closed, can't write");
      return 0;
   }
   switch(ctx->nodeType)
   {
      case DictionaryType:
	 dictionaryAdd(ctx->node,ctx->key,i,ctx->vm);
	 break;
      case ListType:
	 listAppend(ctx->node,i,ctx->vm);
	 break;
   }
   return 1;
}


static int reformat_null(void *ctx) { return append(ctx,Void); }

static int reformat_boolean(void *ctx, int b)
   { return append(ctx,boolCreate(b)); }

static int reformat_int(void *ctx, long long v)
   { return append(ctx,intCreate(v,((ZY_CTX *)ctx)->vm)); }

static int reformat_float(void *ctx, double v)
   { return append(ctx,floatCreate(v,((ZY_CTX *)ctx)->vm)); }

//static int reformat_numStr(void *ctx, const char *s, size_t l) { return 1; }

static int reformat_string(void *_ctx, const unsigned char * stringVal,
                           size_t stringLen)
{ 
   ZY_CTX *ctx = (ZY_CTX *)_ctx;
   return append(ctx,stringCreate2((char *)stringVal,stringLen,ctx->vm));
}

static int reformat_map_key(void *_ctx, const unsigned char *stringVal,
                            size_t stringLen)
{
   ZY_CTX *ctx = (ZY_CTX *)_ctx;
   ctx->key = stringCreate2((char *)stringVal,stringLen,ctx->vm);
   return 1;
}

static int reformat_start_map(void *_ctx)
{
   ZY_CTX   *ctx = (ZY_CTX *)_ctx;
   Instance *m = dictionaryCreate(50,I_OWNED,ctx->vm);

   ctx->stack[ctx->n++] = ctx->node;
   append(_ctx,m);
   ctx->node = m;
   ctx->nodeType = DictionaryType;
   return 1;
}
static int reformat_end_map(void *_ctx)
{
   ZY_CTX *ctx = (ZY_CTX *)_ctx;

   ctx->node = ctx->stack[--ctx->n];
   ctx->nodeType = TYPEO(ctx->node);
   return 1;
}

static int reformat_start_array(void *_ctx)
{
   ZY_CTX   *ctx = (ZY_CTX *)_ctx;
   Instance *a = listCreate(10,0x00,I_OWNED,ctx->vm);

   append(_ctx,a);
   ctx->stack[ctx->n++] = ctx->node;
   ctx->node = a;
   ctx->nodeType = ListType;
   return 1;
}
static int reformat_end_array(void *_ctx)
{
   ZY_CTX *ctx = (ZY_CTX *)_ctx;

   ctx->node = ctx->stack[--ctx->n];
   ctx->nodeType = TYPEO(ctx->node);
   return 1;
}

static yajl_callbacks callbacks = 
{
   reformat_null,
   reformat_boolean,
   reformat_int,
   reformat_float,
   NULL,	// number as string, trumps reformat_int, reformat_float
   reformat_string,
   reformat_start_map,
   reformat_map_key,
   reformat_end_map,
   reformat_start_array,
   reformat_end_array
};

    // .create(vervose=False)
static Instance *yajl_create(Instance *_,pArglist arglist,pVM vm)
{
   ZY_CTX     *ctx = (ZY_CTX *)instanceAllocate(sizeof(ZY_CTX),&yajlObject,1,vm);
   yajl_handle hand    = yajl_alloc(&callbacks,NULL,(void *)ctx);
   Instance   *root    = listCreate(2,0x00,I_OWNED,ctx->vm);
   int	       verbose = arglistTryToGetBool(arglist,0,0,0,vm);
   
   ctx->root = ctx->node = root;
   ctx->nodeType = ListType;
   ctx->src = verbose ? dataCreate(500,I_OWNED,vm) : 0;

    /* and let's allow comments by default */
   ctx->hand = hand;
   yajl_config(hand, yajl_allow_comments, 1);
   yajl_config(hand, yajl_allow_multiple_values, 1);

   addToCollectables((Instance *)ctx,I_OWNED,vm);
   return (Instance *)ctx;
}

    // .close() --> obj
static Instance *yajl_close(Instance *self,pArglist arglist,pVM vm)
{
   ZY_CTX   *ctx  = (ZY_CTX *)self;
   Instance *root;
   int	     s;

   if(ctx->closed) return Void;
   s = yajl_complete_parse(ctx->hand);
   if(s != yajl_status_ok)
   {
      size_t         sz  = 0;
      unsigned char *src = (ctx->src) ? dataText(ctx->src,&sz) : 0;
      unsigned char *msg = yajl_get_error(ctx->hand,src?1:0,src,sz);
      Instance      *tmpCopy;
      sz = strlen((char *)msg);
      if(sz && msg[--sz]=='\n') msg[sz] = '\0';
      tmpCopy = stringCreate((char *)msg,I_OWNED,vm);
      yajl_free_error(ctx->hand,msg);
      yajlFree(self);
      vmThrow(vm,E_VALUE_ERROR,stringText(tmpCopy));
   }
   root = ctx->root;
   root = listGet(root,0);
   yajlFree(self);
   return root ? root : Void;
}

    // .write(text) --> self|throw
static Instance *yajl_write(Instance *self,pArglist arglist,pVM vm)
{
   ZY_CTX *ctx = (ZY_CTX *)self;
   Instance *text = arglistGet(arglist,0,"yajl.write",vm);
   char   *src;
   int	   s;
   size_t  sz;

   s = TYPEO(text);
   if (s==StringType) { src = stringText(text); sz = strlen(src); }
   else if (s==DataType) src = (char *)dataText(text,&sz);
   else arglistGetOnlyString(arglist,0,"yajl.write",vm);  // throw

   if(ctx->closed)
      vmThrow(vm,E_IO_ERROR,"yajl is closed, can't write");

   if(ctx->src) dataInsertBytes(ctx->src,0,(Byte *)src,sz,1,vm);

   ctx->vm = vm;
   s = yajl_parse(ctx->hand,(unsigned char *)src,sz);
   if(s != yajl_status_ok)
   {
//printf("%s\n",yajl_status_to_string(s));
      yajl_close(self,NoArglist,vm);
   }

   return self;
}


static MethodTable yajlMethods[] =
{
   "write",	(pMethod)yajl_write,
   "close",	(pMethod)yajl_close,
   "create",	(pMethod)yajl_create,
   0,		0
};

static Instance *toJSON = 0;  // static immortal class, don't need to mark

DllExport void *construct(void *vm)
{
   extern char wad[];  // wad.c

   if (!vm) return (void *)ZKLX_PROTOCOL;	// handshake
   	// If this is reloaded, nothing happens except 
	// construct() is called again so don't reinitialize
   if (!yajlObject.name)	// static items are zero
   {
      constructObject(&yajlObject,NativeType, yajlMethods,0,0,vm);
      yajlObject.name	      = "YAJL";
      yajlObject.vaultPath    = "Network";
      yajlObject.freeMe       = yajlFree;
      yajlObject.magicMarker  = yajlMarker;
      yajlObject.isize        = sizeof(ZY_CTX);

      toJSON = loadWad((Byte *)wad,0,NoArglist,vm);
      toJSON = classFindFcn(toJSON,"toJSON",0,0,vm);
      if(!toJSON) vmThrow(vm,E_LOADER_ERROR,"toJSON not found");
   }
#if 0
   {
      static ZY_CTX    yajl;
      instanceInit((Instance *)&yajl,&yajlObject,I_UNTOUCHABLE);
      yajl.closed = 1;
      return tupleCreateX(vm,methodCreate((Instance *)&yajl,0,yajl_create,vm),toJSON,ZNIL);
   }
#else
   return tupleCreateX(vm,methodCreate(Void,0,yajl_create,vm),toJSON,ZNIL);
#endif
}


#if 0
        for ( i=1; i < strlen(argv[a]); i++) {
            switch (argv[a][i]) {
                case 'm':
                    yajl_gen_config(g, yajl_gen_beautify, 0);
                    break;
                case 's':
                    yajl_config(hand, yajl_allow_multiple_values, 1);
                    s_streamReformat = 1;
                    break;
                case 'u':
                    yajl_config(hand, yajl_dont_validate_strings, 1);
                    break;
                case 'e':
                    yajl_gen_config(g, yajl_gen_escape_solidus, 1);
                    break;
#endif
